@@ -62,6 +62,14 @@
 
 #define NV_MODESET_MODULE_NAME "nvidia-modeset"
 
+#define NV_VGPU_VFIO_MODULE_NAME "nvidia-vgpu-vfio"
+
+#define NV_NVLINK_MODULE_NAME "nvidia-nvlink"
+#define NV_NVLINK_PROC_PERM_PATH "/proc/driver/nvidia-nvlink/permissions"
+
+#define NV_NVSWITCH_MODULE_NAME "nvidia-nvswitch"
+#define NV_NVSWITCH_PROC_PERM_PATH "/proc/driver/nvidia-nvswitch/permissions"
+
 #define NV_DEVICE_FILE_MODE_MASK (S_IRWXU|S_IRWXG|S_IRWXO)
 #define NV_DEVICE_FILE_MODE (S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH)
 #define NV_DEVICE_FILE_UID 0
@@ -391,18 +399,20 @@ static int modprobe_helper(const int print_errors, const char *module_name)
             return 0;
 
         default:
-            if (waitpid(pid, &status, 0) < 0)
-            {
-                return 0;
-            }
-            if (WIFEXITED(status) && WEXITSTATUS(status) == 0)
-            {
-                return 1;
-            }
-            else
-            {
-                return 0;
-            }
+            /*
+             * waitpid(2) is not always guaranteed to return success even if
+             * the child terminated normally.  For example, if the process
+             * explicitly configured the handling of the SIGCHLD signal
+             * to SIG_IGN, then waitpid(2) will instead block until all
+             * children terminate and return the error ECHILD, regardless
+             * of the child's exit codes.
+             *
+             * Hence, ignore waitpid(2) error codes and instead check
+             * whether the desired kernel module is loaded.
+             */
+            waitpid(pid, NULL, 0);
+
+            return is_kernel_module_loaded(module_name);
     }
 
     return 1;
@@ -477,6 +487,65 @@ static void init_device_file_parameters(uid_t *uid, gid_t *gid, mode_t *mode,
     fclose(fp);
 }
 
+/* 
+ * A helper to query device file states.
+ */
+static int get_file_state_helper(
+    const char *path,
+    int major,
+    int minor,
+    const char *proc_path,
+    uid_t uid,
+    gid_t gid,
+    mode_t mode)
+{
+    dev_t dev = NV_MAKE_DEVICE(major, minor);
+    struct stat stat_buf;
+    int ret;
+    int state = 0;
+
+    ret = stat(path, &stat_buf);
+    if (ret == 0)
+    {
+        nvidia_update_file_state(&state, NvDeviceFileStateFileExists);
+
+        if (S_ISCHR(stat_buf.st_mode) && (stat_buf.st_rdev == dev))
+        {
+            nvidia_update_file_state(&state, NvDeviceFileStateChrDevOk);
+        }
+
+        if (((stat_buf.st_mode & NV_DEVICE_FILE_MODE_MASK) == mode) &&
+            (stat_buf.st_uid == uid) &&
+            (stat_buf.st_gid == gid))
+        {
+            nvidia_update_file_state(&state, NvDeviceFileStatePermissionsOk);
+        }
+    }
+
+    return state;
+}
+
+int nvidia_get_file_state(int minor, int module_instance)
+{
+    char path[NV_MAX_CHARACTER_DEVICE_FILE_STRLEN];
+    char proc_path[NV_MAX_PROC_REGISTRY_PATH_SIZE];
+    mode_t mode;
+    uid_t uid;
+    gid_t gid;
+    int modification_allowed;
+    int state = 0;
+
+    assign_device_file_name(path, minor, module_instance);
+    assign_proc_registry_path(proc_path, module_instance);
+
+    init_device_file_parameters(&uid, &gid, &mode, &modification_allowed,
+                                proc_path);
+
+    state = get_file_state_helper(path, NV_MAJOR_DEVICE_NUMBER, minor,
+                                  proc_path, uid, gid, mode);
+
+    return state;
+}
 
 /*
  * Attempt to create the specified device file with the specified major
@@ -493,7 +562,7 @@ static int mknod_helper(int major, int minor, const char *path,
     gid_t gid;
     int modification_allowed;
     int ret;
-    struct stat stat_buf;
+    int state;
     int do_mknod;
 
     if (path == NULL || path[0] == '\0')
@@ -511,18 +580,12 @@ static int mknod_helper(int major, int minor, const char *path,
         return 1;
     }
 
-    /*
-     * If the device file already exists with correct properties,
-     * nothing to do: success.
-     */
-    ret = stat(path, &stat_buf);
+    state = get_file_state_helper(path, major, minor,
+                                  proc_path, uid, gid, mode);
 
-    if ((ret == 0) &&
-        (S_ISCHR(stat_buf.st_mode)) &&
-        (stat_buf.st_rdev == dev) &&
-        ((stat_buf.st_mode & NV_DEVICE_FILE_MODE_MASK) == mode) &&
-        (stat_buf.st_uid == uid) &&
-        (stat_buf.st_gid == gid))
+    if (nvidia_test_file_state(state, NvDeviceFileStateFileExists) &&
+        nvidia_test_file_state(state, NvDeviceFileStateChrDevOk) &&
+        nvidia_test_file_state(state, NvDeviceFileStatePermissionsOk))
     {
         return 1;
     }
@@ -531,19 +594,18 @@ static int mknod_helper(int major, int minor, const char *path,
 
     do_mknod = 0;
 
-    if (ret != 0)
+    if (!nvidia_test_file_state(state, NvDeviceFileStateFileExists))
     {
         do_mknod = 1;
     }
 
     /*
-     * If the stat(2) above succeeded but the file is either not a
-     * character device or has the wrong major/minor character device
-     * number, then we need to delete it and recreate it.
+     * If the file exists but the file is either not a character device or has
+     * the wrong major/minor character device number, then we need to delete it
+     * and recreate it.
      */
-    if ((ret == 0) &&
-        (!S_ISCHR(stat_buf.st_mode) ||
-         (stat_buf.st_rdev != dev)))
+    if (!do_mknod &&
+        !nvidia_test_file_state(state, NvDeviceFileStateChrDevOk))
     {
         ret = remove(path);
         if (ret != 0)
@@ -714,11 +776,90 @@ int nvidia_modeset_mknod(void)
 {
     char proc_path[NV_MAX_PROC_REGISTRY_PATH_SIZE];
 
-    assign_proc_registry_path(proc_path, 0);
+    assign_proc_registry_path(proc_path, NV_MODULE_INSTANCE_NONE);
 
     return mknod_helper(NV_MAJOR_DEVICE_NUMBER,
                         NV_MODESET_MINOR_DEVICE_NUM,
                         NV_MODESET_DEVICE_NAME, proc_path);
 }
 
+/*
+ * Attempt to create the NVIDIA NVLink driver device file.
+ */
+int nvidia_nvlink_mknod(void)
+{
+    int major = get_chardev_major(NV_NVLINK_MODULE_NAME);
+
+    if (major < 0)
+    {
+        return 0;
+    }
+
+    return mknod_helper(major,
+                        0,
+                        NV_NVLINK_DEVICE_NAME,
+                        NV_NVLINK_PROC_PERM_PATH);
+}
+
+/*
+ * Attempt to create the NVIDIA NVSwitch driver device files.
+ */
+int nvidia_nvswitch_mknod(int minor)
+{
+    int major = 0;
+    char name[NV_MAX_CHARACTER_DEVICE_FILE_STRLEN];
+    int ret;
+
+    major = get_chardev_major(NV_NVSWITCH_MODULE_NAME);
+
+    if (major < 0)
+    {
+        return 0;
+    }
+
+    if (minor == NV_NVSWITCH_CTL_MINOR)
+    {
+        ret = snprintf(name, NV_MAX_CHARACTER_DEVICE_FILE_STRLEN,
+                       NV_NVSWITCH_CTL_NAME);
+    }
+    else
+    {
+        ret = snprintf(name, NV_MAX_CHARACTER_DEVICE_FILE_STRLEN,
+                       NV_NVSWITCH_DEVICE_NAME, minor);
+    }
+
+    if (ret <= 0)
+    {
+        return 0;
+    }
+
+    return mknod_helper(major, minor, name, NV_NVSWITCH_PROC_PERM_PATH);
+}
+
+int nvidia_vgpu_vfio_mknod(int minor_num)
+{
+    int major = get_chardev_major(NV_VGPU_VFIO_MODULE_NAME);
+    char vgpu_dev_name[NV_MAX_CHARACTER_DEVICE_FILE_STRLEN];
+    char proc_path[NV_MAX_PROC_REGISTRY_PATH_SIZE];
+    int ret;
+
+    if (major < 0)
+    {
+        return 0;
+    }
+
+    assign_proc_registry_path(proc_path, NV_MODULE_INSTANCE_NONE);
+
+    ret = snprintf(vgpu_dev_name, NV_MAX_CHARACTER_DEVICE_FILE_STRLEN,
+                   NV_VGPU_VFIO_DEVICE_NAME, minor_num);
+    if (ret <= 0)
+    {
+        return 0;
+    }
+
+    vgpu_dev_name[NV_MAX_CHARACTER_DEVICE_FILE_STRLEN - 1] = '\0';
+
+    return mknod_helper(major, minor_num, vgpu_dev_name, proc_path);
+}
+
 #endif /* NV_LINUX */
